# -*- coding: utf-8 -*-


"""Response rules tests for `whispyr` package"""

import pytest
import uuid

from whispyr import ResponseRule, ClientError


@pytest.mark.parametrize('response_rules', [4], indirect=True)
def test_list_response_rules(workspace, cassette, response_rules):
    _check_list(workspace, response_rules, 4)


@pytest.mark.parametrize('response_rules', [55], indirect=True)
def test_list_response_rules_paginated(workspace, cassette, response_rules):
    _check_list(workspace, response_rules, 55)


def test_show_response_rule(workspace, cassette, response_rule):
    response_rule = workspace.response_rules.show(response_rule['id'])
    _check_basic_response_rule(response_rule)


def test_show_generic_response_rule(whispir, cassette, generic_response_rule):
    response_rule = whispir.response_rules.show(generic_response_rule['id'])
    _check_basic_response_rule(response_rule)


def test_update_reponse_rule(workspace, cassette, response_rule):
    rule_id = response_rule['id']
    name = response_rule['name']
    description = response_rule['description']
    new_name = '{} updated'.format(name)
    new_description = '{} updated'.format(description)
    workspace.response_rules.update(
        rule_id,
        name=new_name,
        description=new_description,
        responseTemplatePatterns=response_rule['responseTemplatePatterns']
    )
    updated_rule = workspace.response_rules.show(rule_id)
    assert updated_rule['name'] == new_name
    assert updated_rule['description'] == new_description


@pytest.mark.parametrize('response_rule', [False], indirect=True)
def test_delete_response_rule(workspace, cassette, response_rule):
    _test_delete(workspace.response_rules, response_rule['id'])


@pytest.mark.parametrize('generic_response_rule', [False], indirect=True)
def test_delete_generic_response_rule(
        whispir, cassette, generic_response_rule):
    _test_delete(whispir.response_rules, generic_response_rule['id'])


def _test_delete(response_rules, rule_id):
    response_rules.delete(rule_id)
    with pytest.raises(ClientError) as excinfo:
        response_rules.show(rule_id)

    exc = excinfo.value
    assert exc.response.status_code == 404


@pytest.fixture(params=[True])
def response_rule(request, workspace):
    delete = request.param
    return _create_fixture_rule(
        request, workspace.response_rules, delete=delete)


@pytest.fixture(params=[True])
def generic_response_rule(request, whispir):
    delete = request.param
    return _create_fixture_rule(
        request, whispir.response_rules, delete=delete)


@pytest.fixture(params=[2])
def response_rules(request, workspace):
    num = request.param
    return [_create_fixture_rule(request, workspace.response_rules)
            for _ in range(num)]


def _create_fixture_rule(request, response_rules, delete=True):
    rule = {
        'name': str(uuid.uuid4()),
        'description': str(uuid.uuid4()),
        'responseTemplatePatterns': {
            'responseTemplatePattern': [{
                'name': str(uuid.uuid4()),
                'textPrompt': 'YES',
                'voicePrompt': '1',
                'spokenVoicePrompt': 'to select YES',
                'pattern': 'startswith',
                'colour': '#00947d'
            }]
        }
    }
    response_rule = response_rules.create(**rule)
    _check_basic_response_rule(response_rule)

    if delete:
        def _delete():
            response_rules.delete(response_rule['id'])

        request.addfinalizer(_delete)

    return response_rule


def _check_list(workspace, response_rules, expected_count):
    rules = workspace.response_rules.list()
    rules = list(rules)
    assert len(rules) == len(response_rules)
    assert len(rules) == expected_count
    for rule in rules:
        _check_basic_response_rule(rule)


def _check_basic_response_rule(response_rule):
    assert isinstance(response_rule, ResponseRule)
    assert 'id' in response_rule
    assert 'name' in response_rule
    assert 'description' in response_rule



# -*- coding: utf-8 -*-

"""Fixtures for `whispyr` tests"""

import base64
import os
import pytest
import prettyserializer
import itertools
import functools
import uuid
import random
import hashlib

from collections import ByteString

from vcr import VCR
from vcr.request import Request
from vcr.util import CaseInsensitiveDict

from whispyr import Whispir


TEST_USERNAME = 'U53RN4M3'
TEST_PASSWORD = 'P4ZZW0RD'
TEST_API_KEY = 'V4L1D4P1K3Y'
TEST_GCM_API_KEY = '9OO9l3ClouDm355491n94P1K3y'


@pytest.fixture
def whispir(pytestconfig):
    username = pytestconfig.getoption('--whispir-username')
    password = pytestconfig.getoption('--whispir-password')
    api_key = pytestconfig.getoption('--whispir-api-key')
    return Whispir(username, password, api_key)


@pytest.fixture
def gcm_api_key(pytestconfig):
    return pytestconfig.getoption('--whispir-gcm-api-key')


@pytest.fixture
def cassette(request, pytestconfig):
    cassette_library_dir = cassettes_dir(request)
    vcr = make_vcr(pytestconfig, cassette_library_dir)
    path = request.function.__name__
    with vcr.use_cassette(path) as cass:
        yield cass


def make_vcr(pytestconfig, cassette_library_dir):
    mode = pytestconfig.getoption('--vcr-mode')
    api_key = pytestconfig.getoption('--whispir-api-key')
    username = pytestconfig.getoption('--whispir-username')
    gcm_api_key = pytestconfig.getoption('--whispir-gcm-api-key')
    scrubber = scrub_patterns(((api_key, TEST_API_KEY),
                               (username, TEST_USERNAME),
                               (gcm_api_key, TEST_GCM_API_KEY)))
    options = {
        'record_mode': mode,
        'filter_headers': [
            ('authorization', replace_auth),
            ('set-cookie', None),
            ('cookie', None)
        ],
        'filter_query_parameters': [('apikey', TEST_API_KEY)],
        'before_record_response': scrubber,
        'before_record_request': scrubber,
        'path_transformer': VCR.ensure_suffix('.yaml'),
        'decode_compressed_response': True,
        'cassette_library_dir': cassette_library_dir,
        'match_on': (
            'method', 'scheme', 'host', 'port', 'path', 'query', 'headers'
        ),
        'serializer': 'pretty-yaml'
    }
    vcr = VCR(**options)
    vcr.register_serializer('pretty-yaml', prettyserializer)
    return vcr


def recorded_fixture(name=None, library_dir='tests/cassettes/fixtures'):
    def decorator(fixture):
        path = name or fixture.__name__

        @pytest.fixture(name=fixture.__name__)
        # TODO: figure out required fixtures from decorated function
        def wrapper(pytestconfig, whispir):
            vcr = make_vcr(pytestconfig, library_dir)
            with vcr.use_cassette(path):
                return fixture(whispir)
        return wrapper

    return decorator


@recorded_fixture()
def workspace(whispir):
    project_name = 'whispyr tests'
    for workspace in whispir.workspaces.list():
        if workspace['projectName'] == project_name:
            return workspace

    workspace = whispir.workspaces.create(projectName=project_name, status='A')
    return whispir.workspaces.show(workspace['id'])


@pytest.fixture(params=[False])
def app(request, whispir, workspace, gcm_api_key, cassette):
    delete = request.param
    return _create_fixture_app(
        request, whispir.apps, workspace, gcm_api_key, delete=delete)


@pytest.fixture(params=[2])
def apps(request, whispir, workspace, gcm_api_key, cassette):
    num = request.param
    return [_create_fixture_app(request, whispir.apps, workspace, gcm_api_key)
            for _ in range(num)]


def _create_fixture_app(request, apps, workspace, gcm_api_key, delete=True):
    name = hashlib.md5(str(uuid.uuid4()).encode('utf-8')).hexdigest()
    app = {
        'name': name,
        'description': 'App ({}) to test whispyr apps CRUD works'.format(name),
        'bundleId': '1',
        'options': {
            'workspaces': workspace['id'],
            'deviceLimit': 3,
            'apiKey': str(uuid.uuid4()),
            'clientSecret': str(uuid.uuid4())
        },
        'gcm': {
            'gcmProjectId': 'foobar-204e3',
            'gcmProjectNumber': 383181709987,
            'gcmApiKey': gcm_api_key
        },
        'registrationTypes': [
            'INVITE'
        ]
    }

    app = apps.create(**app)

    if delete:
        def _delete():
            apps.delete(app['id'])

        request.addfinalizer(_delete)

    return app


@pytest.fixture(params=[False])
def contact(request, workspace, cassette):
    delete = request.param
    return _create_fixture_contact(
        request, workspace.contacts, delete=delete)


@pytest.fixture(params=[True])
def generic_contact(request, whispir, cassette):
    delete = request.param
    return _create_fixture_contact(
        request, whispir.contacts, delete=delete)


@pytest.fixture(params=[2])
def contacts(request, workspace, cassette):
    num = request.param
    return [_create_fixture_contact(request, workspace.contacts)
            for _ in range(num)]


def _create_fixture_contact(request, contacts, delete=True):
    email = 'success+{}@simulator.amazonses.com'.format(uuid.uuid4())
    number = random.randrange(61420000000, 61430000000)
    contact = {
        'firstName': 'John',
        'lastName': 'Wick',
        'status': 'A',
        'timezone': 'Australia/Melbourne',
        'workEmailAddress1': email,
        'workMobilePhone1': str(number),
        'workCountry': 'Australia',
        'messagingoptions': [
            {
                'channel': 'sms',
                'enabled': 'true',
                'primary': 'WorkMobilePhone1'
            },
            {
                'channel': 'email',
                'enabled': 'true',
                'primary': 'WorkEmailAddress1'
            },
            {
                'channel': 'voice',
                'enabled': 'true',
                'primary': 'WorkMobilePhone1'
            }
        ]
    }

    contact = contacts.create(**contact)
    contact = contacts.show(contact['id'])

    if delete:
        def _delete():
            contacts.delete(contact['id'])

        request.addfinalizer(_delete)

    return contact


def replace_auth(key, value, request):
    username = TEST_USERNAME.encode('utf-8')
    password = TEST_PASSWORD.encode('utf-8')
    b64auth = base64.b64encode(b':'.join((username, password)))
    return 'Basic {}'.format(b64auth.decode('utf-8'))


def scrub_pattern(pattern, replacement=''):
    def scrub_it(string):
        return replace(string, pattern, replacement)

    def scrub_them_all(seq):
        return list(map(scrub_it, seq))

    @functools.singledispatch
    def before_record(response):
        if 'Set-Cookie' in response['headers']:
            del response['headers']['Set-Cookie']

        headers = response['headers']
        response['headers'] = {
            name: scrub_them_all(values) for name, values in headers.items()
        }
        new_body = scrub_it(response['body']['string'])
        response['body']['string'] = new_body

        headers = CaseInsensitiveDict(response['headers'])
        if 'content-length' in headers:
            headers['content-length'] = [str(len(new_body))]
            response['headers'] = dict(headers)

        return response

    @before_record.register(Request)
    def _(request):
        orig_body = request.body
        if orig_body:
            new_body = scrub_it(orig_body)
            request.body = new_body
            headers = request.headers
            if 'content-length' in headers:
                headers['content-length'] = str(len(new_body))

        return request

    return before_record


def scrub_patterns(patches):
    pipeline = itertools.starmap(scrub_pattern, patches)
    return compose(pipeline)


def compose(functions):
    def compose_impl(f, g):
        return lambda x: f(g(x))
    return functools.reduce(compose_impl, functions, lambda x: x)


def replace(string, pattern, replacement):
    if isinstance(string, ByteString):
        pattern = pattern.encode('ascii')
        replacement = replacement.encode('ascii')
    return string.replace(pattern, replacement)


def cassettes_dir(request):
    return os.path.join(
        request.fspath.dirname,
        'cassettes',
        request.fspath.basename.replace('.py', ''))


def pytest_addoption(parser):
    parser.addoption('--whispir-username', default=TEST_USERNAME,
                     help='Whispir username')
    parser.addoption('--whispir-password', default=TEST_PASSWORD,
                     help='Whispir password')
    parser.addoption('--whispir-api-key', default=TEST_API_KEY,
                     help='Whispir API key')
    parser.addoption('--whispir-gcm-api-key', default=TEST_GCM_API_KEY,
                     help='Whispir Google Cloud Messaging API key')
    parser.addoption('--vcr-mode', default='once',
                     choices=['once', 'new_episodes', 'none', 'all'],
                     help='Set VCR mode')



# -*- coding: utf-8 -*-

"""Tests for `whispyr` package"""

import base64

import httpretty
from httpretty import HTTPretty

import pytest
import re

import whispyr
from whispyr import ClientError, WhispirRetry

httpretty.HTTPretty.allow_net_connect = False


TEST_USERNAME = 'U53RN4M3'
TEST_PASSWORD = 'P4ZZW0RD'
TEST_API_KEY = 'V4L1D4P1K3Y'


@pytest.fixture(params=[{'retry': WhispirRetry()}])
def whispir(request):
    params = request.param
    return whispyr.Whispir(TEST_USERNAME, TEST_PASSWORD, TEST_API_KEY,
                           **params)


@httpretty.activate
def test_basic_auth(whispir):
    httpretty.register_uri(httpretty.GET, re.compile(r'.*', re.M), body='{}')
    whispir.request('get', 'workspaces')

    assert httpretty.has_request()
    request = httpretty.last_request()
    assert 'Authorization' in request.headers
    basic_auth = request.headers['Authorization']
    expected_basic = _basic_auth(TEST_USERNAME, TEST_PASSWORD)
    assert basic_auth == expected_basic


@httpretty.activate
def test_api_key_provided(whispir):
    httpretty.register_uri(httpretty.GET, re.compile(r'.*', re.M), body='{}')
    whispir.request('get', 'workspaces')

    assert httpretty.has_request()
    request = httpretty.last_request()
    assert 'apikey' in request.querystring
    api_key = request.querystring['apikey'][0]
    assert api_key == TEST_API_KEY


@httpretty.activate
def test_user_agent_is_set(whispir):
    httpretty.register_uri(httpretty.GET, re.compile(r'.*', re.M), body='{}')
    whispir.request('get', 'workspaces')

    assert httpretty.has_request()
    request = httpretty.last_request()
    assert 'User-Agent' in request.headers
    expexted_agent = 'whispyr/{}'.format(whispyr.__version__)
    assert request.headers['User-Agent'] == expexted_agent


@httpretty.activate
def test_retry_succeeded(whispir):
    qps_headers = {
        'X-Mashery-Error-Code': 'ERR_403_DEVELOPER_OVER_QPS',
        'X-Mashery-Error-Detail': 'Account Over Queries Per Second Limit',
        'Retry-After': 1
    }

    qps_response = HTTPretty.Response(
        body='', status=403, adding_headers=qps_headers)

    httpretty.register_uri(
        httpretty.GET, re.compile(r'.*', re.M),
        responses=[
            qps_response,
            qps_response,
            HTTPretty.Response(body='{}', status=200),
        ]
    )

    assert whispir.request('get', 'workspaces') == {}


@httpretty.activate
@pytest.mark.parametrize('whispir', [{'retry': WhispirRetry(total=1)}],
                         indirect=True)
def test_retry_limit(whispir):
    qps_headers = {
        'X-Mashery-Error-Code': 'ERR_403_DEVELOPER_OVER_QPS',
        'X-Mashery-Error-Detail': 'Account Over Queries Per Second Limit',
        'Retry-After': 1
    }

    qps_response = HTTPretty.Response(
        body='', status=403, adding_headers=qps_headers)

    httpretty.register_uri(
        httpretty.GET, re.compile(r'.*', re.M),
        responses=[
            qps_response,
            qps_response,
            qps_response,
            HTTPretty.Response(body='{}', status=200),
        ]
    )

    with pytest.raises(ClientError) as excinfo:
        whispir.request('get', 'workspaces')

    exc = excinfo.value
    assert exc.response.status_code == 403


@httpretty.activate
def test_do_not_retry_qpd(whispir):
    qps_headers = {
        'X-Mashery-Error-Code': 'ERR_403_DEVELOPER_OVER_QPD',
        'X-Mashery-Error-Detail': 'Account Over Queries Per Day Limit',
        'Retry-After': 20 * 60 * 60  # 20 hours in seconds
    }

    httpretty.register_uri(
        httpretty.GET, re.compile(r'.*', re.M),
        responses=[
            HTTPretty.Response(
                body='', status=403, adding_headers=qps_headers),
            HTTPretty.Response(body='{}', status=200),
        ]
    )

    with pytest.raises(ClientError) as excinfo:
        whispir.request('get', 'workspaces')

    exc = excinfo.value
    assert exc.response.status_code == 403


@httpretty.activate
def test_no_retry_for_forbidden(whispir):
    httpretty.register_uri(
        httpretty.GET, re.compile(r'.*', re.M),
        responses=[
            HTTPretty.Response(body='', status=403),
            HTTPretty.Response(body='{}', status=200),
        ]
    )

    with pytest.raises(ClientError) as excinfo:
        whispir.request('get', 'workspaces')

    exc = excinfo.value
    assert exc.response.status_code == 403


def _basic_auth(username, password):
    username = username.encode('latin1')
    password = password.encode('latin1')
    b64auth = base64.b64encode(b':'.join((username, password)))
    return 'Basic {}'.format(b64auth.decode('utf-8'))



# -*- coding: utf-8 -*-

"""Templates tests for `whispyr` package"""

import pytest
import uuid

from whispyr import Template, ClientError


@pytest.fixture(params=[True])
def template(request, workspace):
    delete = request.param
    return _create_fixture_template(
        request, workspace.templates, delete=delete)


@pytest.fixture(params=[True])
def generic_template(request, whispir):
    delete = request.param
    return _create_fixture_template(
        request, whispir.templates, delete=delete)


@pytest.fixture(params=[2])
def templates(workspace, request):
    num = request.param
    templates = []
    for _ in range(num):
        name = str(uuid.uuid4())
        template = _create_fixture_template(request, workspace.templates, name)
        templates.append(template)
    return templates


def _create_fixture_template(request, templates,
                             name='Whispyr test template',
                             description='Create templates test',
                             delete=True):
    template = templates.create(
        messageTemplateName=name,
        messageTemplateDescription=description,
        subject='test template',
        body='This is the body of my test SMS message',
        email={},
        voice={},
        web={}
    )

    if delete:
        def _delete_template():
            templates.delete(template['id'])

        request.addfinalizer(_delete_template)

    return template


@pytest.mark.parametrize('templates', [4], indirect=True)
def test_list_templates(workspace, cassette, templates):
    templates = workspace.templates.list()
    templates = list(templates)
    assert len(templates) == 4
    for template in templates:
        _check_template_instance(template)


@pytest.mark.parametrize('templates', [55], indirect=True)
def test_list_templates_paginated(request, workspace, cassette, templates):
    templates = workspace.templates.list()
    templates = list(templates)
    assert len(templates) == 55
    for template in templates:
        _check_template_instance(template)


def test_show_template(workspace, cassette, template):
    _test_show_template(workspace.templates, template['id'])


def test_show_generic_template(whispir, cassette, generic_template):
    _test_show_template(whispir.templates, generic_template['id'])


def _test_show_template(templates, template_id):
    template = templates.show(template_id)
    _check_template_instance(template)


def test_update_template(workspace, cassette, template):
    description = template['messageTemplateDescription']
    new_description = '{} updated'.format(description)
    workspace.templates.update(
        template['id'],
        messageTemplateName=template['messageTemplateName'],
        messageTemplateDescription=new_description,
        subject=template['subject'],
        body=template['body'])
    updated_template = workspace.templates.show(template['id'])
    _check_template_instance(updated_template)
    assert updated_template['messageTemplateDescription'] == new_description


@pytest.mark.parametrize('template', [False], indirect=True)
def test_delete_template(workspace, cassette, template):
    _test_delete(workspace.templates, template)


@pytest.mark.parametrize('generic_template', [False], indirect=True)
def test_delete_generic_template(whispir, cassette, generic_template):
    _test_delete(whispir.templates, generic_template)


def _test_delete(templates, template):
    templates.delete(template['id'])
    with pytest.raises(ClientError) as excinfo:
        templates.show(template['id'])

    exc = excinfo.value
    assert exc.response.status_code == 404


def _check_template_instance(template):
    assert isinstance(template, Template)
    assert 'id' in template
    assert 'messageTemplateName' in template
    assert 'messageTemplateDescription' in template



# -*- coding: utf-8 -*-

"""YAML serializer"""

import yaml

# Use the libYAML versions if possible
try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper


def deserialize(cassette_string):
    return yaml.load(cassette_string, Loader=Loader)


class PrettyDumper(Dumper):
    def increase_indent(self, flow=False, indentless=False):
        """this function indent collections to get this:

        foo:
          - bar
        quz:
          - baz

        instead of default:

        foo:
        - bar
        quz:
        - baz

        """
        return super(PrettyDumper, self).increase_indent(flow, False)


def literal_string_representer(dumper, data):
    return dumper.represent_scalar('tag:yaml.org,2002:str', data, style='|')


class literal_string(str):
    pass


def serialize(cassette_dict):
    for interaction in cassette_dict['interactions']:
        response = interaction['response']
        response_body = response['body']
        if response_body is not None:
            string = response_body.get('string', None)
            if string:
                response_body['string'] = literal_string(string)

        request = interaction['request']
        request_body = request['body']
        if request_body is not None:
            request['body'] = literal_string(request_body)
    yaml.add_representer(literal_string, literal_string_representer)
    return yaml.dump(
        cassette_dict, Dumper=PrettyDumper, default_flow_style=False)



# -*- coding: utf-8 -*-


"""Contacts tests for `whispyr` package"""

import pytest

from whispyr import Contact, ClientError


@pytest.mark.parametrize('contacts', [4], indirect=True)
def test_list_contacts(workspace, cassette, contacts):
    _check_list(workspace, contacts, 4)


@pytest.mark.parametrize('contacts', [55], indirect=True)
def test_list_contacts_paginated(workspace, cassette, contacts):
    _check_list(workspace, contacts, 55)


def test_show_contact(workspace, cassette, contact):
    contact = workspace.contacts.show(contact['id'])
    _check_basic_contact(contact)


def test_show_generic_contact(whispir, cassette, generic_contact):
    contact = whispir.contacts.show(generic_contact['id'])
    _check_basic_contact(contact)


def test_update_contact(workspace, cassette, contact):
    contact_id = contact['id']
    first_name = contact['firstName']
    last_name = contact['lastName']
    new_first_name = '{} updated'.format(first_name)
    new_last_name = '{} updated'.format(last_name)
    workspace.contacts.update(
        contact_id,
        firstName=new_first_name,
        lastName=new_last_name,
        timezone=contact['timezone'],
        workCountry=contact['workCountry']
    )
    updated_contact = workspace.contacts.show(contact_id)
    assert updated_contact['firstName'] == new_first_name
    assert updated_contact['lastName'] == new_last_name


@pytest.mark.parametrize('contact', [False], indirect=True)
def test_delete_contact(workspace, cassette, contact):
    _test_delete(workspace.contacts, contact['id'])


@pytest.mark.parametrize('generic_contact', [False], indirect=True)
def test_delete_generic_contact(whispir, cassette, generic_contact):
    _test_delete(whispir.contacts, generic_contact['id'])


def _test_delete(contacts, contact_id):
    contacts.delete(contact_id)
    with pytest.raises(ClientError) as excinfo:
        contacts.show(contact_id)

    exc = excinfo.value
    assert exc.response.status_code == 404


def _check_list(workspace, contacts, expected_count):
    contacts = workspace.contacts.list()
    contacts = list(contacts)
    assert len(contacts) == len(contacts)
    assert len(contacts) == expected_count
    for contact in contacts:
        _check_basic_contact(contact)


def _check_basic_contact(contact):
    assert isinstance(contact, Contact)
    assert 'id' in contact
    assert 'firstName' in contact
    assert 'lastName' in contact



# -*- coding: utf-8 -*-


"""Tests for `whispyr` package"""

import itertools

from urllib.parse import urlparse

from whispyr import Message, MessageStatus, MessageResponse


def test_messages_create(workspace, cassette):
    message = workspace.messages.create(
        to='success@simulator.amazonses.com',
        subject='whispyr test',
        body='test message, please disregard')
    _check_created_message(message, cassette)


def test_send_generic_message(whispir, cassette):
    message = whispir.messages.create(
        to='success@simulator.amazonses.com',
        subject='whispyr test',
        body='test message, please disregard')
    _check_created_message(message, cassette)


def test_send_alias_creates_message(workspace, cassette):
    message = workspace.messages.send(
        to='success@simulator.amazonses.com',
        subject='whispyr test',
        body='test message, please disregard')
    _check_created_message(message, cassette)


def test_show_message(workspace, cassette):
    _test_show_message(workspace.messages)


def test_show_generic_message(whispir, cassette):
    _test_show_message(whispir.messages)


def _test_show_message(messages):
    message = next(messages.list())
    message = messages.show(message['id'])
    _check_message(message)


def test_list_message_statuses(workspace, cassette):
    # message ID here is explicitly specified as it's easier than to
    # send a response back from test
    message = workspace.messages.Message(id='09CD9065116DF39F')
    statuses = message.statuses.list()
    statuses = list(statuses)
    assert len(statuses) > 0
    for status in statuses:
        assert isinstance(status, MessageStatus)
        assert 'categories' in status


def test_list_message_responses(workspace, cassette):
    # message ID here is explicitly specified as it's easier than to
    # send a response back from test
    message = workspace.messages.Message(id='09CD9065116DF39F')
    responses = message.responses.list()
    responses = list(responses)
    assert len(responses) > 0
    for response in responses:
        assert isinstance(response, MessageResponse)
        assert 'responseCount' in response


def _check_message(message):
    assert isinstance(message, Message)
    assert 'id' in message
    assert 'subject' in message


def test_list_generic_messages(whispir, cassette):
    messages = whispir.messages.list()
    _check_list_messages(messages)


def test_list_messages(workspace, cassette):
    messages = workspace.messages.list()
    _check_list_messages(messages)


def _check_list_messages(messages):
    messages = list(itertools.islice(messages, 25))
    assert len(messages) > 0
    for message in messages:
        assert isinstance(message, Message)
        assert 'id' in message
        assert 'subject' in message


def _check_created_message(message, cassette):
    assert len(cassette) == 1
    location = cassette.responses[-1]['headers']['Location'][0]
    assert message['id'] == _msg_location(location)
    assert isinstance(message, Message)


def _msg_location(url):
    path = urlparse(url).path
    return path.split('/')[-1]



# -*- coding: utf-8 -*-

"""Main module."""

from . import __version__

from collections import UserDict

from requests import Session
from requests.adapters import HTTPAdapter
from requests.auth import HTTPBasicAuth, AuthBase

from urllib.parse import urljoin, urlparse, parse_qsl

from urllib3.util import Retry
from urllib3.exceptions import MaxRetryError

WHISPIR_BASE_URL = 'https://api.whispir.com'


class WhispirError(Exception):

    def __init__(self, response):
        super().__init__()
        self.response = response


class ClientError(WhispirError):
    pass


class ServerError(WhispirError):
    pass


class JSONDecodeError(WhispirError):
    pass


class WhispirAuth(AuthBase):

    def __init__(self, api_key, username, password):
        self._api_key = api_key
        self._basic_auth = HTTPBasicAuth(username, password)

    def __call__(self, request):
        new_req = self._basic_auth(request)
        new_req.prepare_url(new_req.url, {'apikey': self._api_key})
        return new_req


class WhispirRetry(Retry):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.RETRY_AFTER_STATUS_CODES = frozenset([403, 413, 429, 503])
        self.raise_on_status = False
        self.raise_on_redirect = False

    def _is_method_retryable(self, method):
        return True

    def increment(self, method=None, url=None, response=None, error=None,
                  _pool=None, _stacktrace=None):
        if response:
            mashery_error = response.getheader("X-Mashery-Error-Code")
            if mashery_error == 'ERR_403_DEVELOPER_OVER_QPD':
                raise MaxRetryError(_pool, url, error)
        return super().increment(
            method=method, url=url, response=response, error=error,
            _pool=_pool, _stacktrace=_stacktrace)


DEFAULT_RETRY = WhispirRetry()


class Whispir:

    def __init__(self, username, password, api_key,
                 base_url=WHISPIR_BASE_URL, page_size=20, retry=DEFAULT_RETRY):
        self._base_url = base_url
        self.page_size = page_size
        self._session = Session()
        self._session.auth = WhispirAuth(api_key, username, password)
        adapter = HTTPAdapter(max_retries=retry)
        self._session.mount('http://', adapter)
        self._session.mount('https://', adapter)
        self._session.headers.update({
            'User-Agent': 'whispyr/{}'.format(__version__)
        })
        # collections
        self.workspaces = Workspaces(self)
        self.messages = Messages(self)
        self.templates = Templates(self)
        self.response_rules = ResponseRules(self)
        self.contacts = Contacts(self)
        self.apps = Apps(self)

    def request(self, method, path, **kwargs):
        url = urljoin(self._base_url, path)
        response = self._session.request(method, url, **kwargs)
        if response.ok:
            return self._maybe_return_json(response)
        else:
            if 400 <= response.status_code < 500:
                error = ClientError
            elif 500 <= self.status_code < 600:
                error = ServerError

            raise error(response)

    def _maybe_return_json(self, response):
        if not response.content:
            return

        try:
            return response.json()
        except ValueError:
            raise JSONDecodeError(response)


class Collection:

    def __init__(self, whispir, base_container=None):
        self.whispir = whispir
        class_name = self.__class__.__name__
        self.name = (getattr(self, 'name', False) or class_name.lower())
        type_name = (getattr(self, 'type_name', False) or
                     _singularize(self.name))
        self.vnd_type = 'application/vnd.whispir.{}-v1+json'.format(type_name)
        self.list_name = getattr(self, 'list_name', self.name)
        self.container = (getattr(self, 'container', False) or
                          globals()[_singularize(class_name)])
        self.resource = (getattr(self, 'resource', False) or self.name)
        self.base_container = base_container

        class ContainerProxy(self.container):
            collection = self
            container = self.container

            def __new__(self, *args, **kwargs):
                return self.container(self.collection, *args, **kwargs)

        setattr(self, self.container.__name__, ContainerProxy)

    def path(self, id=None):
        path = self.resource
        if id:
            path = '{}/{}'.format(path, id)

        if self.base_container:
            path = '{}/{}'.format(self.base_container.path(), path)

        return path

    def request(self, method, path, headers=None, **kwargs):
        headers = kwargs.pop('headers', {})
        collection_type = self.vnd_type
        headers.update({
            'Content-Type': collection_type,
            'Accept': collection_type
        })
        return self.whispir.request(method, path, headers=headers, **kwargs)

    def _containerize(self, item):
        return self.container(self, **item)

    def create(self, **kwargs):
        path = self.path()
        item = self.request('post', path, json=kwargs)
        return self._containerize(item)

    def show(self, id):
        path = self.path(id)
        item = self.request('get', path)
        return self._containerize(item)

    def _get_page(self, path, **kwargs):
        result = self._try_get(path, kwargs)
        return self._page_items(result)

    def _page_items(self, response):
        return response.get(self.list_name, [])

    def list(self, **kwargs):
        path = self.path()

        list_items = self._list
        if 'offset' in kwargs or 'limit' in kwargs:
            list_items = self._get_page

        for item in list_items(path, **kwargs):
            yield self._containerize(item)

    def _list(self, path, **kwargs):
        kwargs['limit'] = self.whispir.page_size
        kwargs['offset'] = 0

        while True:
            result = self._try_get(path, kwargs)
            items = self._page_items(result)
            for item in items:
                yield item

            links = result.get('link', [])
            link = _find_link(links, 'next')
            if not link:
                return

            uri = urlparse(link['uri'])
            query = dict(parse_qsl(uri.query))
            kwargs['limit'] = query['limit']
            kwargs['offset'] = query['offset']

    def _try_get(self, path, params):
        try:
            return self.request('get', path, params=params)
        except (ClientError, JSONDecodeError) as e:
            if e.response.status_code == 404:
                return {}
            raise

    def update(self, id, **kwargs):
        path = self.path(id)
        self.request('put', path, json=kwargs)

    def delete(self, id):
        path = self.path(id)
        self.request('delete', path)


class Nonpaginatable:

    def _list(self, path, **kwargs):
        return self._get_page(path, **kwargs)


class Streamable:

    def _list(self, path, **kwargs):
        limit = self.whispir.page_size
        kwargs['limit'] = limit
        kwargs['offset'] = 0

        while True:
            item = None
            for item in self._get_page(path, **kwargs):
                yield item

            if not item:
                break

            kwargs['offset'] += limit


class Container(UserDict):

    def __init__(self, collection, id=None, **kwargs):
        self.collection = collection
        self.whispir = collection.whispir

        if not id:
            id = self.id_from_links(kwargs.get('link', []))

        if id:
            kwargs['id'] = id

        super().__init__(kwargs)

    @classmethod
    def id_from_links(cls, links):
        link = _find_link(links, 'self')
        if link:
            return cls.id_from_uri(link['uri'])

    @staticmethod
    def id_from_uri(url):
        path = urlparse(url).path
        return path.split('/')[-1]

    def path(self):
        return self.collection.path(self.id())

    def id(self):
        return self['id']


class Workspaces(Nonpaginatable, Collection):
    pass


class Messages(Streamable, Collection):

    def create(self, **kwargs):
        try:
            super().create(**kwargs)
        except JSONDecodeError as e:
            headers = e.response.headers
            msg_id = self.Message.id_from_uri(headers['location'])

        return self.Message(id=msg_id)

    send = create


class MessageStatuses(Collection):

    list_name = 'messageStatuses'
    resource = 'messagestatus'


class MessageResponses(Collection):
    pass


class Templates(Collection):
    list_name = 'messagetemplates'


class ResponseRules(Nonpaginatable, Collection):
    list_name = 'responseRules'


class Contacts(Collection):
    pass


class Apps(Collection):
    list_name = 'applications'


class Workspace(Container):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.messages = Messages(self.whispir, self)
        self.templates = Templates(self.whispir, self)
        self.response_rules = ResponseRules(self.whispir, self)
        self.contacts = Contacts(self.whispir, self)
        self.apps = Apps(self.whispir, self)


class Message(Container):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.statuses = MessageStatuses(self.whispir, self)
        self.responses = MessageResponses(self.whispir, self)


class MessageStatus(Container):
    pass


class MessageResponse(Container):
    pass


class Template(Container):
    pass


class ResponseRule(Container):
    pass


class Contact(Container):
    pass


class App(Container):
    pass


def _singularize(string):
    rules = [('ies', 'y'), ('uses', 'us'), ('s', '')]
    for suffix, replacement in rules:
        if string.endswith(suffix):
            return string[:-len(suffix)] + replacement

    return string


def _find_link(links, relation, default=None):
    def is_relation(it):
        return it['rel'] == relation
    return next((it for it in links if is_relation(it)), default)



# -*- coding: utf-8 -*-

"""Tests for `whispyr` workspaces"""

from whispyr import Workspace


def test_create_workspace(cassette, whispir):
    workspace = whispir.workspaces.create(
        projectName='whispyr tests', status='A')
    assert isinstance(workspace, Workspace)
    assert 'id' in workspace


def test_list_workspaces(cassette, whispir):
    workspaces = whispir.workspaces.list()
    workspaces = list(workspaces)
    assert len(workspaces) > 0
    for workspace in workspaces:
        _check_workspace(workspace)


def test_show_workspace(whispir, workspace, cassette):
    workspace = whispir.workspaces.show(workspace['id'])
    _check_workspace(workspace)


def _check_workspace(workspace):
    assert isinstance(workspace, Workspace)
    assert 'id' in workspace
    assert 'projectName' in workspace



# -*- coding: utf-8 -*-

"""Top-level package for whispyr."""

__author__ = """Grigory Starinkin"""
__email__ = 'starinkin@gmail.com'
__version__ = '0.1.0'

from .whispyr import Whispir, WhispirRetry

from .whispyr import Message, MessageStatus, MessageResponse, Template, \
    Workspace, ResponseRule, Contact, App

from .whispyr import WhispirError, ClientError, ServerError, JSONDecodeError

__all__ = [
    # Client
    'Whispir', 'WhispirRetry',
    # Resources
    'Message', 'MessageStatus', 'MessageResponse', 'Template', 'Workspace',
    'ResponseRule', 'Contact', 'App',
    # Errors
    'WhispirError', 'ClientError', 'ServerError', 'JSONDecodeError'
]



# -*- coding: utf-8 -*-


"""Apps tests for `whispyr` package"""

import pytest

from whispyr import App, ClientError


@pytest.mark.parametrize('apps', [4], indirect=True)
def test_list_apps(whispir, cassette, apps):
    _check_list(whispir, apps, 4)


@pytest.mark.parametrize('apps', [55], indirect=True)
def test_list_apps_paginated(whispir, cassette, apps):
    _check_list(whispir, apps, 55)


def test_show_app(whispir, cassette, app):
    app = whispir.apps.show(app['id'])
    _check_basic_app(app)


def test_update_app(whispir, cassette, app):
    app_id = app['id']
    name = app['name']
    description = app['description']
    new_name = '{} updated'.format(name)
    new_description = '{} updated'.format(description)
    whispir.apps.update(
        app_id,
        name=new_name,
        description=new_description,
        bundleId=app['bundleId']
    )
    updated_app = whispir.apps.show(app_id)
    assert updated_app['name'] == new_name
    assert updated_app['description'] == new_description


@pytest.mark.parametrize('app', [False], indirect=True)
def test_delete_app(whispir, cassette, app):
    _test_delete(whispir.apps, app['id'])


def _test_delete(apps, app_id):
    apps.delete(app_id)
    with pytest.raises(ClientError) as excinfo:
        apps.show(app_id)

    exc = excinfo.value
    assert exc.response.status_code == 404


def _check_list(workspace, apps, expected_count):
    apps = workspace.apps.list()
    apps = list(apps)
    assert len(apps) == len(apps)
    assert len(apps) == expected_count
    for app in apps:
        _check_basic_app(app)


def _check_basic_app(app):
    assert isinstance(app, App)
    assert 'id' in app
    assert 'name' in app
    assert 'description' in app
